package jgame;
import java.awt.*;
import java.util.*;
import java.awt.image.*;
import java.net.*;

/** Some handy utilities for loading an manipulating images; used internally
 * by jgame. */
public class ImageUtil {

	Component output_comp;
	public MediaTracker mediatracker;
	public DummyObserver observer = new DummyObserver();

	Hashtable loadedimages = new Hashtable(); /* filenames => Images */

	class DummyObserver implements ImageObserver {
		public DummyObserver() {}
		public boolean imageUpdate(Image img, int infoflags,
		int x,int y, int width,int height) { return true; }
	}

	public ImageUtil(Component comp) {
		output_comp=comp;
		mediatracker = new MediaTracker(output_comp);
	}

	/** Returns null if there was an error.
	public Image loadImageFromURL(String imgurl) {
		Image img = (Image)loadedimages.get(imgurl);
		if (img==null) {
			img = output_comp.getToolkit().createImage(imgurl);
			loadedimages.put(imgurl,img);
		}
		try {
			ensureLoaded(img);
		} catch (Exception e) {
			System.err.println("Error loading image "+imgurl);
			return null;
		}
		return img;
	} */

	/** Load image from resource path (using getResource).
	* Returns null if there was an error. */
	public Image loadImage(String imgfile) {
		Image img = (Image)loadedimages.get(imgfile);
		if (img==null) {
			URL imgurl = getClass().getClassLoader().getResource(imgfile);
			img = output_comp.getToolkit().createImage(imgurl);
			loadedimages.put(imgfile,img);
		}
		try {
			ensureLoaded(img);
		} catch (Exception e) {
			System.err.println("Error loading image "+imgfile);
			return null;
		}
		return img;
	}

	/** Returns null if there was an error. */
	public Image loadImage(URL imgurl) {
		Image img = (Image)loadedimages.get(imgurl);
		if (img==null) {
			img = output_comp.getToolkit().createImage(imgurl);
			loadedimages.put(imgurl,img);
		}
		try {
			ensureLoaded(img);
		} catch (Exception e) {
			System.err.println("Error loading image "+imgurl);
			return null;
		}
		return img;
	}

	public Dimension getSize(Image img) {
		return new Dimension(
			img.getWidth(observer),
			img.getHeight(observer) );
	}

	/** true means the image has some transparent pixels, false means image
	* is completely opaque. It actually checks the alpha channel pixel for
	* pixel.  */
	public boolean isOpaque(Image img) {
		//System.out.print("isOpaque start"+System.currentTimeMillis());
		BufferedImage bimg = toBuffered(img);
		WritableRaster ras = bimg.getAlphaRaster();
		int [] alpha = ras.getPixels(ras.getMinX(), ras.getMinY(),
			ras.getWidth(), ras.getHeight(), (int []) null);
		//for (int i=0; i<alpha.length; i++)
		//	System.out.print(alpha[i]);
		for (int i=0; i<alpha.length; i++) if (alpha[i]!=1) {
			//System.out.println("False");
			return false;
		}
		//System.out.println("True");
		return true;
	}

	public int [] getPixels(Image img) {
		Dimension size = getSize(img);
		int [] buffer = new int [size.width * size.height];
		PixelGrabber grabber = new PixelGrabber(img, 0, 0,
				size.width,size.height, buffer, 0, size.width);
		grabPixels(grabber);
		return buffer;
	}

	public int [] getPixels(Image img, int x,int y,int width,int height) {
		int [] buffer = new int [width * height];
		PixelGrabber grabber = new PixelGrabber(img, x, y,
				width,height, buffer, 0, width);
		grabPixels(grabber);
		return buffer;
	}

	private void grabPixels(PixelGrabber grabber) {
		try {
			grabber.grabPixels();
		} catch(InterruptedException e) {
			e.printStackTrace();
		}
	}

	/** for angle, only increments of 90 are allowed */
	public Image rotate(Image img,int angle) {
		Image rot = null;
		Dimension size = getSize(img);
		int [] buffer = getPixels(img);
		int [] rotate = new int [size.width * size.height];
		int angletype = (angle/90) & 3;
		if (angletype==0) return img;
		if (angletype==1) {
			/* 1 2 3 4    9 5 1
			 * 5 6 7 8 => a 6 2
			 * 9 a b c    b 7 3 
			 *            c 8 4 */
			for(int y = 0; y < size.height; y++) {
				for(int x = 0; x < size.width; x++) {
					rotate[x*size.height + (size.height-1-y) ] =
							buffer[(y*size.width)+x];
				}
			}
			return output_comp.createImage(
				new MemoryImageSource(size.height,size.width,
					rotate, 0, size.height) );
		}
		if (angletype==3) {
			/* 1 2 3 4    4 8 c
			 * 5 6 7 8 => 3 7 b
			 * 9 a b c    2 6 a 
			 *            1 5 9 */
			for(int y = 0; y < size.height; y++) {
				for(int x = 0; x < size.width; x++) {
					rotate[(size.width-x-1)*size.height + y] =
							buffer[(y*size.width)+x];
				}
			}
			return output_comp.createImage(
				new MemoryImageSource(size.height,size.width,
					rotate, 0, size.height) );
		}
		if (angletype==2) {
			/* 1 2 3 4    c b a 9
			 * 5 6 7 8 => 8 7 6 5
			 * 9 a b c    4 3 2 1 */
			for(int y = 0; y < size.height; y++) {
				for(int x = 0; x < size.width; x++) {
					rotate[((size.height-y-1)*size.width)+(size.width-x-1)] =
							buffer[(y*size.width)+x];
				}
			}
		}
		return output_comp.createImage(
			new MemoryImageSource(size.width,size.height,
				rotate, 0, size.width) );
	}

	public Image flip(Image img,boolean horiz,boolean vert) {
		Image flipped = null;
		Dimension size = getSize(img);
		int [] buffer = getPixels(img);
		int [] flipbuf = new int [size.width * size.height];
		if (vert && !horiz) {
			for(int y = 0; y < size.height; y++) {
				for(int x = 0; x < size.width; x++) {
					flipbuf[(size.height-y-1)*size.width + x] =
							buffer[(y*size.width)+x];
				}
			}
		} else if (horiz && !vert) {
			for(int y = 0; y < size.height; y++) {
				for(int x = 0; x < size.width; x++) {
					flipbuf[y*size.width + (size.width-x-1)] =
							buffer[(y*size.width)+x];
				}
			}
		} else if (horiz && vert) {
			for(int y = 0; y < size.height; y++) {
				for(int x = 0; x < size.width; x++) {
					flipbuf[(size.height-y-1)*size.width + (size.width-x-1)] =
							buffer[(y*size.width)+x];
				}
			}
		}
		return output_comp.createImage(
			new MemoryImageSource(size.width,size.height,
				flipbuf, 0, size.width) );
	}
	public Image scale(Image img, int width, int height) {
		//BufferedImage dstimg = createCompatibleImage(width,height);
		//BufferedImage srcimg = toBuffered(img);
		Image scaledimg=img.getScaledInstance(width,height,Image.SCALE_SMOOTH);
		try {
			/* this is necessary */
			ensureLoaded(scaledimg);
		} catch (Exception e) {
			System.err.println("Error scaling image.");
		}
		return scaledimg;
	}
	public void ensureLoaded(Image img) throws Exception {
		//System.err.println("In ensureloaded");
		mediatracker.addImage(img,0);
		try {
			mediatracker.waitForAll();
			if (mediatracker.getErrorsAny()!=null) {
				mediatracker.removeImage(img);
				throw new Exception("Error loading image");
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		mediatracker.removeImage(img);
		//System.err.println("Out ensureloaded");
	}

	public Image crop(Image img, int x,int y, int width,int height) {
		Dimension size = getSize(img);
		int [] buffer = getPixels(img,x,y, width,height);
		return output_comp.createImage(
			new MemoryImageSource(width,height,
				buffer, 0, width) );
	}

	/** Create BufferedImage version of image for faster drawing; returns the
	* identical object if it was already a BufferedImage.  For X11, you
	* have to do this so that you get pixmaps instead of ximages. Only handles
	* bitmask-type transparency.
	*/
	public BufferedImage toBuffered(Image img) {
		if (img instanceof BufferedImage) return (BufferedImage)img;
		/* see the developer's almanac,
		* http://javaalmanac.com/egs/java.awt.image/Image2Buf.html,
		* for docs */
		Dimension size = getSize(img);
		BufferedImage bimage = createCompatibleImage(size.width,size.height);
		Graphics g = bimage.createGraphics();
		g.drawImage(img, 0, 0, null);
		g.dispose();
		return bimage;
	}

	/** Create empty bitmask image that should be efficient on this display */
	private BufferedImage createCompatibleImage(int width,int height) {
		//try {
		GraphicsEnvironment ge =
				GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gs = ge.getDefaultScreenDevice();
		GraphicsConfiguration gc = gs.getDefaultConfiguration();
		// always use bitmask transparency
		BufferedImage bimage = gc.createCompatibleImage(
				width, height, Transparency.BITMASK);
		return bimage;
		//} catch (HeadlessException e) { // this exception is not in 1.2
			// The system does not have a screen
		//	e.printStackTrace();
		//	return null;
		//}
	}

}

